/*
 [The "BSD licence"]
 Copyright (c) 2007, John Snyders
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package jjs.stst;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.ResourceBundle;

import org.antlr.stringtemplate.AutoIndentWriter;
import org.antlr.stringtemplate.NoIndentWriter;
import org.antlr.stringtemplate.PathGroupLoader;
import org.antlr.stringtemplate.StringTemplate;
import org.antlr.stringtemplate.StringTemplateErrorListener;
import org.antlr.stringtemplate.StringTemplateGroup;
import org.antlr.stringtemplate.StringTemplateWriter;
import org.antlr.stringtemplate.language.AngleBracketTemplateLexer;

import org.json.JSONException;
import org.json.JSONObject;

import antlr.RecognitionException;
import antlr.TokenStreamRecognitionException;

/**
 * StringTemplate is both a language for describing text generation and a tool (template
 * engine) that implements that language and provides an API to connect the tool to another
 * program.
 * 
 * The StringTemplate language by design has no way to perform computation or evaluation.
 * It also has no general way to specify structural data. It is up to some driving program
 * (sometimes known as the controler) to produce and or make data available to the StringTemplate
 * template engine. The lack of readily available data makes learning the template language
 * difficult without also learning the StringTemplate API. 
 * 
 * In theory one benifit of the separation of presentation from business logic is that
 * a project can be divided between people who work on the business logic and people who
 * work on the presentation. The people working on presentation don't need to be programmers.
 * 
 * Tutorials on StringTemplate tend to describe both the template syntax and how to invoke 
 * it from the Java API at the same time. In some cases it is confusing to learn 
 * two things at once and more importantly it excludes non-programmers from learning
 * StringTemplate.
 * 
 * In contrast, trying out XSLT is is easy (even if learning XSLT is hard) because all you need
 * is a text editor to create both your input documents and your stylesheet and a XSLT program
 * to process the stylesheet with the input documents.  
 * 
 * The intent of this STStandaloneTool is to make StringTemplate easy to try out without
 * having to program in Java (or some other language). It does this by providing a textual
 * data format that can be input to the tool along with a set of templates. 
 * 
 * What is needed to go along with StringTemplate is a format for representing data. XSLT
 * uses XML as its data model. Another text format that is just as expressive is JSON.
 * 
 * This command line program takes as input a text file in JSON format, the name of a 
 * template and optional group and produces an output file that results from processing the
 * named template with the data from the JSON file.
 * 
 * The command line syntax is
 * stst [<options>] [<group-name>.]<template-name> [<json-file>]
 *    options:
 *    -h                  display usage/help
 *    -n                  no indent
 *    -l                  lint
 *    -d                  debug
 *    -v                  verbose
 *    -a                  use angle bracket lexer
 *    -e <encoding>       encoding for templates, json-file, and output
 *    -o <file>           output file if not specified use stdout
 *    -t <dir>            if not specified the current working directory is used
 * 
 * See usage for most up to date syntax
 * 
 * The STStandaloneTool class can be used for both simple template files (.st extension) and 
 * group files (.stg extension). 
 * 
 * TODO
 *  - support renderers
 *  - automate build version and build number
 *  - independent encoding control for template, data, output
 *  - take some options from ENV VAR?
 *  - more testing: more error handeling, encodings, multiple dirs
 * 
 * future: 
 *  interactive mode
 *  support more data formats: XML, CSV, YAML
 *  print diagnostic info
 *  servlet
 * 
 * @author John Snyders
 *
 */
public class STStandaloneTool
{
    private static final String VERSION = "0.1";
    private static final String RESOURCE_BUNDLE_NAME = "jjs.stst.ApplicationMessages";
    private static final String GROUP_NAME = "main"; // name of group when using simple templates

    private static ResourceBundle resources = null;

    {
        resources = ResourceBundle.getBundle(RESOURCE_BUNDLE_NAME);
    }

    // group to hold templates when using simple templates or 
    // read group file into when using a group file
    private StringTemplateGroup group = null;

    // a place to report errors
    private STErrorListener errorListener = null;

    //options
    private boolean noIndent = false;
    private boolean lintMode = false;
    private boolean debugMode = false;
    private boolean useAngleBracketLexer = false;

    // where to write template output
    private File outFile = null;

    // the JSON data used as input to the template
    private JSONObject data = null;

    public STStandaloneTool()
    {
        errorListener = new STErrorListener();
    }

    /**
     * Controls the output indenting
     * @return true if not indenting output and false otherwise
     */
    public boolean getNoIndent()
    {
        return noIndent;
    }

    /**
     * Controls the output indenting
     * @param noIndent true to not indent and false to indent output
     */
    public void setNoIndent(boolean noIndent)
    {
        this.noIndent = noIndent;
    }

    /**
     * Controls lint mode.
     * @return true if lint mode enabled false otherwise
     */
    public boolean isLintMode()
    {
        return lintMode;
    }

    /**
     * Controls lint mode.
     * @param lintMode true to turn lint mode on and false to turn it off
     */
    public void setLintMode(boolean lintMode)
    {
        this.lintMode = lintMode;
    }

    /**
     * Controls the lexer to use. < > or $ $
     * @return true if using angle brackets and false for dollar signs
     */
    public boolean getUseAngleBracketLexer()
    {
        return useAngleBracketLexer;
    }

    /**
     * Controls the lexer to use. < > or $ $
     * @param useAngleBracketLexer true for angle brackets and false for dollar signs
     */
    public void setUseAngleBracketLexer(boolean useAngleBracketLexer)
    {
        this.useAngleBracketLexer = useAngleBracketLexer;
    }

    /**
     * Controls debug mode. In debug mode start end "tags" will mark template boundaries
     * in the output.
     * @return true if in debug mode false otherwise
     */
    public boolean isDebugMode()
    {
        return debugMode;
    }

    /**
     * Controls debug mode. In debug mode start end "tags" will mark template boundaries
     * in the output.
     * @param debugMode true for debug mode and false for no debugging
     */
    public void setDebugMode(boolean debugMode)
    {
        this.debugMode = debugMode;
    }

    /**
     * The file to write template output to. If null then write output to standard output.
     * @return output file or null if none
     */
    public File getOutFile()
    {
        return outFile;
    }

    /**
     * The file to write template output to. If null then write output to standard output.
     * @param outFile output file or null to use standard output
     */
    public void setOutFile(File outFile)
    {
        this.outFile = outFile;
    }

    /**
     * The data the template will use
     * @return JSON internal representation of template input data
     */
    public JSONObject getData()
    {
        return data;
    }

    /**
     * Set template data with JSONObject
     * @param data template input data
     */
    public void setData(JSONObject data)
    {
        this.data = data;
    }

    /**
     * Set template data with a string in JSON format
     * @param jsonString template input data string
     */
    public void setData(String jsonString)
    {
        JSONObject data = null;
        try
        {
            data = new JSONObject(jsonString);
        }
        catch (JSONException je)
        {
            String format = resources.getString("JSONError");
            logError(MessageFormat.format(format, je.getLocalizedMessage()));
            throw new RuntimeException(); // exit
        }
        setData(data);
    }

    /**
     * Set template data from the contents of a file
     * @param f the file containing JSON format data
     * @param encoding the file encoding
     */
    public void setData(File f, String encoding)
    {
        String contents = null;
        try
        {
            contents = FileUtil.getFileContentAsString(f, encoding);
        }
        catch (FileNotFoundException fnfe)
        {
            String format = resources.getString("DataFileNotFound");
            logError(MessageFormat.format(format, f.getPath()));
            throw new RuntimeException(); // exit
        }
        catch (IOException ioe)
        {
            String format = resources.getString("ErrorReadingData");
            logError(MessageFormat.format(format, ioe.getLocalizedMessage()));
            throw new RuntimeException(); // exit
        }
        setData(contents);
    }

    /**
     * Set template data from a stream
     * @param is the stream containing JSON format data
     * @param encoding the stream encoding
     */
    public void setData(InputStream is, String encoding)
    {
        String contents = null;
        try
        {
            contents = FileUtil.getStreamAsString(is, encoding);
        }
        catch (IOException ioe)
        {
            String format = resources.getString("ErrorReadingData");
            logError(MessageFormat.format(format, ioe.getLocalizedMessage()));
            throw new RuntimeException(); // exit
        }
        setData(contents);
    }

    /**
     * Only use this method when using a group file. This specifies the directories
     * to search for group files and the encoding for any files.
     * 
     * Note: only one of setGroupLoader, setGroup(name) or setGroup(dir, encoding) should be called
     * 
     * @param dirs a ":" separated list of paths 
     * @param encoding the file encoding for group and related files
     */
    public void setGroupLoader(String dirs, String encoding)
    {
        PathGroupLoader loader = new PathGroupLoader(dirs, errorListener);
        loader.setFileCharEncoding(encoding);
        StringTemplateGroup.registerGroupLoader(loader);

        if(getUseAngleBracketLexer())
        {
            StringTemplateGroup.registerDefaultLexer(AngleBracketTemplateLexer.class);
        }
    }

    /**
     * Set the group name of the main group when using a group file.
     * The setGroupLoader must be called first.
     * 
     * @param groupName name of the group file to process without the extension
     */
    public void setGroup(String groupName)
    {
        group = StringTemplateGroup.loadGroup(groupName);
        if (group == null)
        {
            // an error has already been given
            throw new RuntimeException(); // exit
        }
        initGroup();
    }

    /**
     * Set the group when using simple template files
     * All the template files must be under the given directory
     * 
     * Note: only one of setGroupLoader, setGroup(name) or setGroup(dir, encoding) should be called
     * 
     * @param dir path directory containing templates
     * @param encoding encoding for any and all template files if null the 
     * system default encoding is used
     */
    public void setGroup(String dir, String encoding)
    {
        if(getUseAngleBracketLexer())
        {
            group = new StringTemplateGroup(GROUP_NAME, dir, AngleBracketTemplateLexer.class);
        }
        else
        {
            group = new StringTemplateGroup(GROUP_NAME, dir);
        }
        if (encoding != null)
        {
            group.setFileCharEncoding(encoding);
        }
        initGroup();
    }

    /**
     * Call to generate output from the given template.
     * Call after setting desired options and after setting the data and group
     * 
     * @param templateName the template to process
     */
    public void generate(String templateName)
    {
        StringTemplate.setLintMode(isLintMode());
        StringTemplate st = getTemplate(templateName);
        invokeTemplate(st);
    }

    /**
     * List the instance settings
     */
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        sb.append("StringTemplateStandaloneTool:\n");
        sb.append("Group: ").append(group.getName()).append("\n");
        sb.append("Indent: ").append(noIndent ? "yes" : "no").append("\n");
        sb.append("Lint: ").append(lintMode ? "yes" : "no").append("\n");
        sb.append("Debug: ").append(debugMode ? "yes" : "no").append("\n");
        sb.append("Lexer: ").append(useAngleBracketLexer ? "angle brackets" : "dollar sign").append("\n");
        sb.append("Output: ").append(outFile != null ? outFile.getName() : "<stdout>").append("\n");
        return sb.toString();
    }
    //
    // Implementation
    //

    private void initGroup()
    {
        group.setErrorListener(errorListener);
        group.registerRenderer(String.class, new BasicFormatRenderer());
        // TODO 
        // support extensible renderers

        group.emitDebugStartStopStrings(isDebugMode());
    }

    private StringTemplate getTemplate(String templateName)
    {
        StringTemplate st = null;
        try
        {
            st = group.getInstanceOf(templateName);
        }
        catch (IllegalArgumentException iae)
        {
            // I would think that getInstanceOf should call the error
            // listener but it doesn't
            String msg = resources.getString("NoSuchTemplate");
            errorListener.error(msg, iae);
            throw new RuntimeException(); // exit
        }
        return st;
    }

    private void invokeTemplate(StringTemplate st)
    {
        st.setAttributes(getData());

        PrintWriter out = null;
        try
        {
            out = getOutputPrintWriter();
            StringTemplateWriter writer = null;
            if (getNoIndent())
            {
                writer = new NoIndentWriter(out);
            }
            else
            {
                writer = new AutoIndentWriter(out);
            }
            st.write(writer);
            out.println();
            out.flush();
        }
        catch (Exception ex)
        {
            String msg = resources.getString("RuntimeError");
            errorListener.error(msg, ex);
            throw new RuntimeException(); // exit
        }
        finally
        {
            if (outFile != null)
            {
                out.close();
            }
        }
    }

    private PrintWriter getOutputPrintWriter()
    {
        if (outFile != null)
        {
            try
            {
                return new PrintWriter(outFile);
            }
            catch (FileNotFoundException ex)
            {
                String format = resources.getString("OutputFileNotFound");
                logError(MessageFormat.format(format, outFile.getPath()));
                throw new RuntimeException(); // exit
            }
        }
        return new PrintWriter(System.out, true);
    }

    private class STErrorListener implements StringTemplateErrorListener
    {
        public void error(String msg, Throwable e)
        {
            if (e == null)
            {
                String format = resources.getString("Error");
                logError(MessageFormat.format(format, msg));
            }
            else if (e instanceof TokenStreamRecognitionException)
            {
                TokenStreamRecognitionException tsr = (TokenStreamRecognitionException)e;
                RecognitionException r = tsr.recog;
                String format = resources.getString("RecognitionExceptionError");
                logError(MessageFormat.format(format, msg, e.getLocalizedMessage(), 
                        r.getLine(), r.getColumn()));
            }
            else if (e instanceof RecognitionException)
            {
                RecognitionException r = (RecognitionException)e;
                String format = resources.getString("RecognitionExceptionError");
                logError(MessageFormat.format(format, msg, e.getLocalizedMessage(), 
                        r.getLine(), r.getColumn()));
            }
            else
            {
                String format = resources.getString("ExceptionError");
                logError(MessageFormat.format(format, msg, e.getLocalizedMessage()));
                
            }
        }

        public void warning(String msg)
        {
            String format = resources.getString("Warning");
            logError(MessageFormat.format(format, msg));
        }
    }

    private void logError(String message)
    {
        System.err.println(message);
    }

    /**
     * Parse the StringTemplateTool command line and execute it
     * 
     * @param args see usage
     */
    public static void main(String[] args)
    {
        STStandaloneTool stst = new STStandaloneTool();
        String templateSpec = null;
        String data = null;
        ArrayList<File> dirs = new ArrayList<File>();
        String encoding = System.getProperty("file.encoding");
        boolean dirsParam = false;
        boolean outParam = false;
        boolean encodingParam = false;
        boolean verbose = false;

        int param = 0;
        for (String arg : args)
        {
            if (dirsParam)
            {
                dirsParam = false;
                File d = new File(arg);
                if (!d.isDirectory())
                {
                    String format = resources.getString("InvalidDirectory");
                    System.err.println(MessageFormat.format(format, arg));
                    continue;
                }
                dirs.add(d);
            }
            else if (outParam)
            {
                outParam = false;
                stst.setOutFile(new File(arg));
            }
            else if (encodingParam)
            {
                encodingParam = false;
                encoding = arg;
            }
            else if (arg.charAt(0) == '-') // handle options
            {
                if (arg.equals("-n"))
                {
                    stst.setNoIndent(true);
                }
                else if (arg.equals("-l"))
                {
                    stst.setLintMode(true);
                }
                else if (arg.equals("-d"))
                {
                    stst.setDebugMode(true);
                }
                else if (arg.equals("-a"))
                {
                    stst.setUseAngleBracketLexer(true);
                }
                else if (arg.equals("-v"))
                {
                    verbose = true;
                }
                else if (arg.equals("-h"))
                {
                    usage();
                    return;
                }
                else if (arg.equals("-t"))
                {
                    dirsParam = true;
                }
                else if (arg.equals("-o"))
                {
                    outParam = true;
                }
                else if (arg.equals("-e"))
                {
                    encodingParam = true;
                }
                else
                {
                    String format = resources.getString("UnknownOption");
                    System.err.println(MessageFormat.format(format, arg));
                }
            }
            else
            {
                switch (param)
                {
                case 0: // this is the template name
                {
                    templateSpec = arg;
                    break;
                }
                case 1:
                {
                    data = arg;
                    break;
                }
                default:
                {
                    String format = resources.getString("TooManyParameters");
                    System.err.println(MessageFormat.format(format, arg));
                    break;
                }
                }
                param++;
            }
        }
        
        if (param < 1)
        {
            String msg = resources.getString("TooFewParameters");
            System.err.println(msg);
            usage();
            return;
        }

        if (verbose)
        {
            versionBanner();
        }

        // add cwd to list of dirs
        File cwd = new File(System.getProperty("user.dir"));
        dirs.add(cwd);

        try
        {
            String templateName = null;
            // check for group in templateSpec
            int dot = templateSpec.indexOf('.');
            if (dot != -1)
            {
                String groupName = templateSpec.substring(0, dot);
                templateName = templateSpec.substring(dot + 1);
    
                // add all directories to group loader
                StringBuilder sb = new StringBuilder();
                int i = 0;
                for (File dir : dirs)
                {
                    if (i > 0)
                    {
                        sb.append(":");
                    }
                    sb.append(dir.getPath());
                    i++;
                }
                stst.setGroupLoader(sb.toString(), encoding);
                stst.setGroup(groupName);
            }
            else
            {
                stst.setGroup(dirs.get(0).getPath(), encoding);
                templateName = templateSpec;
            }
            
            if (data != null)
            {
                stst.setData(new File(data), encoding);
            }
            else
            {
                stst.setData(System.in, encoding);
            }

            if (verbose)
            {
                String settings = stst.toString();
                System.out.println(settings);
            }
            long start = System.currentTimeMillis();
            stst.generate(templateName);
            long end = System.currentTimeMillis();
            if (verbose)
            {
                double time = (double)(end - start) / 1000.0;
                String format = resources.getString("Timing");
                System.out.println(MessageFormat.format(format, String.valueOf(time)));
            }
        }
        catch (RuntimeException ex)
        {
            // the error was already logged 
            System.exit(1);
        }
    }

    public static void versionBanner()
    {
        String format = resources.getString("VersionBanner");
        System.out.println(MessageFormat.format(format, VERSION, StringTemplate.VERSION));
    }

    public static void usage()
    {
        versionBanner();
        String msg = resources.getString("Usage");
        System.out.println(msg);
    }

}
